use std::{collections::HashMap, fmt, time::Instant};

use crate::PreLispValue;

#[cfg(test)]
mod tests {
	use crate::{LispState, LispValue, Parser, Tokenizer};

	use super::add_function;

	fn get_exec_result(source: &str) -> Result<Vec<LispValue>, Box<dyn std::error::Error>> {
		let mut state = LispState::new();

		state
			.table
			.insert(String::from("output"), LispValue::List(vec![], false));
		add_function(&mut state.table, "print", |st, args| {
			if let LispValue::List(v, _) = st.table.get_mut("output").unwrap() {
				v.push(LispValue::List(args, false))
			}

			Ok(vec![LispValue::Nil])
		});

		let mut tokenizer = Tokenizer::new(String::from(source));
		let mut parser = Parser::new(tokenizer.tokenize()?);
		let instructions = parser.parse();

		state.execute(instructions)?;

		let mut out = Vec::new();
		if let LispValue::List(outerlist, _) = &state.table["output"] {
			for list in outerlist {
				if let LispValue::List(innerlist, _) = list {
					for innervalue in innerlist {
						out.push(innervalue.to_owned());
					}
				}
			}
		}

		Ok(out)
	}

	#[test]
	fn test_hello_world() -> Result<(), Box<dyn std::error::Error>> {
		assert_eq!(
			vec![LispValue::String(String::from("Hello, World!"))],
			get_exec_result(
				"
				(print \"Hello, World!\")
			"
			)?
		);
		Ok(())
	}

	#[test]
	fn test_math() -> Result<(), Box<dyn std::error::Error>> {
		assert_eq!(
			vec![LispValue::Float(16.05)],
			get_exec_result(
				"
				(print 
					(add 
						(div
							4.5
							(sub 0.5 0.2)
						)
						(mul 21 0.05)
					)
				)
			"
			)?
		);
		Ok(())
	}
}

pub struct LispState {
	table: HashMap<String, LispValue>,
}

#[derive(Debug)]
pub struct RuntimeError {
	message: String,
}

impl fmt::Display for RuntimeError {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		write!(f, "Runtime Error: {}", self.message)
	}
}

impl std::error::Error for RuntimeError {}

type RustFunctionType = fn(&mut LispState, Vec<LispValue>) -> Result<Vec<LispValue>, RuntimeError>;

#[derive(Debug, Clone, PartialEq)]
pub enum LispValue {
	Nil,
	True,
	False,
	String(String),
	Integer(i32),
	Float(f32),
	List(Vec<LispValue>, bool),
	Instant(Instant),
	LispFunction(String, Vec<String>, Vec<LispValue>),
	RustFunction(String, RustFunctionType),
	Map(HashMap<String, LispValue>),
	Var(String),
}

impl fmt::Display for LispValue {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		#[allow(unreachable_patterns)]
		match self {
			LispValue::Nil => write!(f, "nil"),
			LispValue::True => write!(f, "true"),
			LispValue::False => write!(f, "false"),
			LispValue::String(str) => write!(f, "\"{}\"", str),
			LispValue::LispFunction(name, _, _) => write!(f, "<function {}>", name),
			LispValue::RustFunction(name, _) => write!(f, "<function {}>", name),
			LispValue::Integer(num) => write!(f, "{}", num),
			LispValue::Float(num) => write!(f, "{}", num),
			LispValue::List(elems, quoted) => {
				let mut strs = Vec::new();

				for e in elems {
					strs.push(e.to_string());
				}

				if *quoted {
					write!(f, "'({})", strs.join(" "))
				} else {
					write!(f, "({})", strs.join(" "))
				}
			}
			val => {
				write!(f, "{}", format!("{:?}", val))
			}
		}
	}
}

fn lisp_display(v: LispValue) -> String {
	#[allow(unreachable_patterns)]
	match v {
		LispValue::Nil => String::from("nil"),
		LispValue::True => String::from("true"),
		LispValue::False => String::from("false"),
		LispValue::String(str) => str,
		LispValue::LispFunction(name, _, _) => format!("<function {}>", name),
		LispValue::RustFunction(name, _) => format!("<function {}>", name),
		LispValue::Integer(num) => num.to_string(),
		LispValue::Float(num) => num.to_string(),
		LispValue::List(elems, quoted) => {
			let mut strs = Vec::new();

			for e in elems {
				strs.push(e.to_string());
			}

			if quoted {
				format!("'({})", strs.join(" "))
			} else {
				format!("({})", strs.join(" "))
			}
		}
		val => {
			format!("{}", val)
		}
	}
}

fn get_type(v: &LispValue) -> String {
	match v {
		LispValue::Nil => String::from("nil"),
		LispValue::True => String::from("bool"),
		LispValue::False => String::from("bool"),
		LispValue::Integer(_) => String::from("int"),
		LispValue::Float(_) => String::from("float"),
		LispValue::String(_) => String::from("string"),
		LispValue::List(_, _) => String::from("list"),
		LispValue::Instant(_) => String::from("instant"),
		LispValue::LispFunction(_, _, _) => String::from("function"),
		LispValue::RustFunction(_, _) => String::from("function"),
		LispValue::Map(_) => String::from("map"),
		LispValue::Var(_) => String::from("symbol"),
	}
}

fn is_falsy(v: &LispValue) -> bool {
	match v {
		LispValue::Nil => true,
		LispValue::False => true,
		_ => false,
	}
}

fn is_truthy(v: &LispValue) -> bool {
	!is_falsy(v)
}

impl From<&str> for LispValue {
	fn from(value: &str) -> Self {
		LispValue::String(value.to_string())
	}
}
impl From<String> for LispValue {
	fn from(value: String) -> Self {
		LispValue::String(value)
	}
}
impl From<i32> for LispValue {
	fn from(value: i32) -> Self {
		LispValue::Integer(value)
	}
}
impl From<f32> for LispValue {
	fn from(value: f32) -> Self {
		LispValue::Float(value)
	}
}
impl From<bool> for LispValue {
	fn from(value: bool) -> Self {
		match value {
			true => LispValue::True,
			false => LispValue::False,
		}
	}
}
impl From<Vec<LispValue>> for LispValue {
	fn from(value: Vec<LispValue>) -> Self {
		LispValue::List(value, false)
	}
}

fn add_function<T: Into<String> + Clone>(
	table: &mut HashMap<String, LispValue>,
	name: T,
	func: RustFunctionType,
) {
	table.insert(
		name.clone().into(),
		LispValue::RustFunction(name.into(), func),
	);
}

impl LispState {
	pub fn new() -> LispState {
		let mut table = HashMap::new();

		add_function(&mut table, "now", |_, _| {
			Ok(vec![LispValue::Instant(Instant::now())])
		});

		add_function(&mut table, "since", |_, args| {
			if args.len() != 1 {
				return Err(RuntimeError {
					message: format!("Bad arguments to since (expected 1, got {})", args.len()),
				});
			}
			if let LispValue::Instant(i) = args[0] {
				Ok(vec![LispValue::Float(
					Instant::now().duration_since(i).as_secs_f32(),
				)])
			} else {
				Err(RuntimeError {
					message: format!(
						"Bad arguments to since (expected instant, got {})",
						get_type(&args[0])
					),
				})
			}
		});

		add_function(&mut table, "type", |_, args| {
			if args.len() != 1 {
				return Err(RuntimeError {
					message: format!("Bad arguments to type (expected 1, got {})", args.len()),
				});
			}
			Ok(vec![LispValue::String(get_type(&args[0]))])
		});

		add_function(&mut table, "sleep", |_, args| {
			if args.len() != 1 {
				return Err(RuntimeError {
					message: format!("Bad arguments to sleep (expected 1, got {})", args.len()),
				});
			}

			if let LispValue::Float(n) = args[0] {
				std::thread::sleep(std::time::Duration::from_secs_f32(n));
			} else if let LispValue::Integer(n) = args[0] {
				std::thread::sleep(std::time::Duration::from_secs(n as u64));
			} else {
				return Err(RuntimeError {
					message: format!(
						"Bad arguments to sleep (expected float or int, got {})",
						get_type(&args[0])
					),
				});
			}

			Ok(vec![LispValue::Nil])
		});

		add_function(&mut table, "or", |_, args| {
			if args.len() != 2 {
				return Err(RuntimeError {
					message: format!("Bad arguments to or (expected 2, got {})", args.len()),
				});
			}
			if is_truthy(&args[0]) {
				Ok(vec![args[0].clone()])
			} else {
				Ok(vec![args[1].clone()])
			}
		});

		add_function(&mut table, "and", |_, args| {
			if args.len() != 2 {
				return Err(RuntimeError {
					message: format!("Bad arguments to and (expected 2, got {})", args.len()),
				});
			}
			if is_truthy(&args[0]) {
				Ok(vec![args[1].clone()])
			} else {
				Ok(vec![args[0].clone()])
			}
		});

		add_function(&mut table, "echo", |_, args| {
			return Ok(args);
		});

		add_function(&mut table, "set", |state, args| {
			if args.len() != 2 {
				return Err(RuntimeError {
					message: format!("Bad arguments to set (expected 2, got {})", args.len()),
				});
			}

			if let LispValue::String(name) = &args[0] {
				state.table.insert(name.to_owned(), args[1].clone());
			} else {
				return Err(RuntimeError {
					message: format!(
						"Bad argument 1 to set (expected string, got {})",
						get_type(&args[0])
					),
				});
			}

			Ok(vec![args[1].clone()])
		});

		add_function(&mut table, "get", |_, args| {
			if args.len() != 2 {
				return Err(RuntimeError {
					message: format!("Bad arguments to get (expected 2, got {})", args.len()),
				});
			}

			if let LispValue::Map(map) = &args[0] {
				if let LispValue::String(str) = &args[1] {
					return Ok(vec![map.get(str).unwrap_or(&LispValue::Nil).to_owned()]);
				} else {
					return Err(RuntimeError {
						message: format!(
							"Bad argument 2 to get (expected string, got {})",
							get_type(&args[0])
						),
					});
				}
			} else {
				return Err(RuntimeError {
					message: format!(
						"Bad argument 1 to get (expected map, got {})",
						get_type(&args[0])
					),
				});
			}
		});

		add_function(&mut table, "print", |_, args| {
			let mut strings = Vec::new();
			for val in args {
				strings.push(lisp_display(val));
			}

			let str = strings.join(" ");
			println!("{}", str);

			Ok(vec![LispValue::Nil])
		});

		add_function(&mut table, "concat", |_, args| {
			let mut strings = Vec::new();
			for val in args {
				strings.push(lisp_display(val));
			}

			let str = strings.concat();
			Ok(vec![LispValue::String(str)])
		});

		add_function(&mut table, "add", |_, args| {
			if args.len() < 1 {
				return Ok(vec![LispValue::Nil]);
			} else if args.len() == 1 {
				return Ok(vec![args[0].clone()]);
			}

			let mut t = "int";
			for v in &args {
				match v {
					LispValue::Float(_) => t = "float",
					LispValue::Integer(_) => (),
					_ => {
						return Err(RuntimeError {
							message: format!(
								"Invalid argument to 'add' (expected float or int, got {}",
								get_type(v)
							),
						})
					}
				}
			}

			match t {
				"int" => {
					let mut sum = 0;
					for v in &args {
						if let LispValue::Integer(n) = v {
							sum += n;
						} else {
							unreachable!()
						}
					}
					return Ok(vec![LispValue::Integer(sum)]);
				}
				"float" => {
					let mut sum = 0.0 as f32;
					for v in &args {
						if let LispValue::Float(n) = v {
							sum += n;
						} else if let LispValue::Integer(n) = v {
							sum += *n as f32;
						} else {
							unreachable!()
						}
					}
					return Ok(vec![LispValue::Float(sum)]);
				}
				_ => unreachable!(),
			}
		});

		add_function(&mut table, "sub", |_, args| {
			if args.len() < 1 {
				return Ok(vec![LispValue::Nil]);
			} else if args.len() == 1 {
				return Ok(vec![args[0].clone()]);
			}

			let mut t = "int";
			for v in &args {
				match v {
					LispValue::Float(_) => t = "float",
					LispValue::Integer(_) => (),
					_ => {
						return Err(RuntimeError {
							message: format!(
								"Invalid argument to 'sub' (expected float or int, got {}",
								get_type(v)
							),
						})
					}
				}
			}

			match t {
				"int" => {
					if let LispValue::Integer(n) = args[0] {
						let mut sum = n;
						for i in 1..args.len() {
							if let LispValue::Integer(n) = &args[i] {
								sum -= n;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Integer(sum)]);
					} else {
						unreachable!()
					}
				}
				"float" => {
					if let LispValue::Float(n) = args[0] {
						let mut sum = n;
						for i in 1..args.len() {
							if let LispValue::Float(n) = &args[i] {
								sum -= *n;
							} else if let LispValue::Integer(n) = &args[i] {
								sum -= *n as f32;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Float(sum)]);
					} else if let LispValue::Integer(n) = args[0] {
						let mut sum = n as f32;
						for i in 1..args.len() {
							if let LispValue::Float(n) = &args[i] {
								sum -= *n;
							} else if let LispValue::Integer(n) = &args[i] {
								sum -= *n as f32;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Float(sum)]);
					}
					{
						unreachable!()
					}
				}
				_ => unreachable!(),
			}
		});

		add_function(&mut table, "mul", |_, args| {
			if args.len() < 1 {
				return Ok(vec![LispValue::Nil]);
			} else if args.len() == 1 {
				return Ok(vec![args[0].clone()]);
			}

			let mut t = "int";
			for v in &args {
				match v {
					LispValue::Float(_) => t = "float",
					LispValue::Integer(_) => (),
					_ => {
						return Err(RuntimeError {
							message: format!(
								"Invalid argument to 'mul' (expected float or int, got {}",
								get_type(v)
							),
						})
					}
				}
			}

			match t {
				"int" => {
					if let LispValue::Integer(n) = args[0] {
						let mut sum = n;
						for i in 1..args.len() {
							if let LispValue::Integer(n) = &args[i] {
								sum *= n;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Integer(sum)]);
					} else {
						unreachable!()
					}
				}
				"float" => {
					if let LispValue::Float(n) = args[0] {
						let mut product = n;
						for i in 1..args.len() {
							if let LispValue::Float(n) = &args[i] {
								product *= *n;
							} else if let LispValue::Integer(n) = &args[i] {
								product *= *n as f32;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Float(product)]);
					} else if let LispValue::Integer(n) = args[0] {
						let mut product = n as f32;
						for i in 1..args.len() {
							if let LispValue::Float(n) = &args[i] {
								product *= *n;
							} else if let LispValue::Integer(n) = &args[i] {
								product *= *n as f32;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Float(product)]);
					}
					{
						unreachable!()
					}
				}
				_ => unreachable!(),
			}
		});

		add_function(&mut table, "div", |_, args| {
			if args.len() < 1 {
				return Ok(vec![LispValue::Nil]);
			} else if args.len() == 1 {
				return Ok(vec![args[0].clone()]);
			}

			let mut t = "int";
			for v in &args {
				match v {
					LispValue::Float(_) => t = "float",
					LispValue::Integer(_) => (),
					_ => {
						return Err(RuntimeError {
							message: format!(
								"Invalid argument to 'div' (expected float or int, got {}",
								get_type(v)
							),
						})
					}
				}
			}

			match t {
				"int" => {
					if let LispValue::Integer(n) = args[0] {
						let mut sum = n;
						for i in 1..args.len() {
							if let LispValue::Integer(n) = &args[i] {
								sum /= n;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Integer(sum)]);
					} else {
						unreachable!()
					}
				}
				"float" => {
					if let LispValue::Float(n) = args[0] {
						let mut product = n;
						for i in 1..args.len() {
							if let LispValue::Float(n) = &args[i] {
								product /= *n;
							} else if let LispValue::Integer(n) = &args[i] {
								product /= *n as f32;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Float(product)]);
					} else if let LispValue::Integer(n) = args[0] {
						let mut product = n as f32;
						for i in 1..args.len() {
							if let LispValue::Float(n) = &args[i] {
								product /= *n;
							} else if let LispValue::Integer(n) = &args[i] {
								product /= *n as f32;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Float(product)]);
					}
					{
						unreachable!()
					}
				}
				_ => unreachable!(),
			}
		});

		LispState { table }
	}

	fn handle_prevalues(&self, input: Vec<PreLispValue>) -> Vec<LispValue> {
		let mut out = Vec::new();

		for v in input {
			out.push(match v {
				PreLispValue::Nil => LispValue::Nil,
				PreLispValue::True => LispValue::True,
				PreLispValue::False => LispValue::False,
				PreLispValue::String(str) => LispValue::String(str),
				PreLispValue::Integer(num) => LispValue::Integer(num),
				PreLispValue::Float(num) => LispValue::Float(num),
				PreLispValue::List(list, quoted) => {
					LispValue::List(Self::handle_prevalues(self, list), quoted)
				}
				PreLispValue::LispFunction(name, arg_names, instructions) => {
					LispValue::LispFunction(
						name,
						arg_names,
						Self::handle_prevalues(self, instructions),
					)
				}
				PreLispValue::Var(name) => LispValue::Var(name),
			})
		}

		out
	}

	fn eval_list(
		&mut self,
		env: &mut HashMap<String, LispValue>,
		list: &Vec<LispValue>,
		quoted: bool,
	) -> Result<(Vec<LispValue>, bool), RuntimeError> {
		if list.len() < 1 {
			return Ok((vec![LispValue::List(Vec::new(), quoted)], false));
		}

		let list = &mut list.clone();

		for i in 0..list.len() {
			if let LispValue::Var(name) = &list[i] {
				list[i] = env
					.get(name)
					.unwrap_or_else(|| self.table.get(name).unwrap_or(&LispValue::Nil))
					.to_owned();
			}
			if let LispValue::List(elements, quoted) = &list[i] {
				let (elems, astuple) = self.eval_list(env, elements, *quoted)?;
				if astuple {
					list[i] = elems.get(0).unwrap_or(&LispValue::Nil).to_owned();
					for j in 1..elems.len() {
						list.insert(i, elems[j].to_owned());
					}
				} else {
					list[i] = LispValue::List(elems, *quoted);
				}
			}
		}

		if quoted {
			return Ok((list.to_vec(), false));
		}

		if let LispValue::RustFunction(_name, f) = &list[0] {
			Ok((f(self, list[1..].to_vec())?, true))
		} else if let LispValue::LispFunction(_name, _, _) = &list[0] {
			Ok((
				self.call_lisp_func_internal(env, &list[0], list[1..].to_vec())?,
				true,
			))
		} else {
			Ok((list.to_vec(), false))
		}
	}

	fn call_lisp_func_internal(
		&mut self,
		env: &mut HashMap<String, LispValue>,
		f: &LispValue,
		args: Vec<LispValue>,
	) -> Result<Vec<LispValue>, RuntimeError> {
		let (arg_names, instructions) = match f {
			LispValue::LispFunction(_, arg_names, i) => (arg_names, i),
			_ => {
				return Err(RuntimeError {
					message: format!(
						"Invalid argument to call_lisp_func_internal (expected function, got {})",
						get_type(f)
					),
				})
			}
		};

		for i in 0..arg_names.len() {
			env.insert(
				arg_names[i].to_owned(),
				args.get(i).unwrap_or(&LispValue::Nil).to_owned(),
			);
		}

		let mut out = Vec::new();
		for l in instructions {
			if let LispValue::List(elements, quoted) = l {
				let (values, astuple) = self.eval_list(env, &elements, *quoted)?;
				if values.len() > 0 && values[0] != LispValue::Nil {
					if astuple {
						for v in values {
							out.push(v);
						}
					} else {
						out.push(values.into());
					}
				}
			}
		}
		Ok(out)
	}

	pub fn call_lisp_func(
		&mut self,
		f: &LispValue,
		args: Vec<impl Into<LispValue>>,
	) -> Result<Vec<LispValue>, RuntimeError> {
		if !matches!(f, LispValue::LispFunction(_, _, _)) {
			return Err(RuntimeError {
				message: format!(
					"Invalid argument to call_lisp_func_internal (expected function, got {})",
					get_type(&f)
				),
			});
		}

		let mut vargs = Vec::new();
		for v in args {
			vargs.push(v.into());
		}

		self.call_lisp_func_internal(&mut HashMap::new(), &f, vargs)
	}

	pub fn get_func(&mut self, name: impl Into<String>) -> Option<LispValue> {
		if let Some(v) = self.table.get(&name.into()) {
			Some(v.to_owned())
		} else {
			None
		}
	}

	pub fn execute(
		&mut self,
		instructions: Vec<PreLispValue>,
	) -> Result<Vec<LispValue>, RuntimeError> {
		let instructions = Self::handle_prevalues(self, instructions);

		let mut outs = Vec::new();

		let mut env = HashMap::new();

		for val in instructions {
			if let LispValue::List(elements, quoted) = val {
				let (vals, astuple) = self.eval_list(&mut env, &elements, quoted)?;
				if astuple {
					for val in vals {
						outs.push(val);
					}
				} else {
					outs.push(vals.into())
				}
			} else {
				outs.push(val);
			}
		}

		Ok(outs)
	}
}
